{% extends '!base.html' %}
{%- block site_meta -%}
  {{ super() }}
  <script type="application/ld+json">
    {
      "@context": "https://schema.org",
      "@type": "TechArticle",
      "headline": "{{ docstitle }}",
      "name": "{{ project }}",
      "description": "Niquests is an elegant and simple HTTP library for Python, built for human beings. It is designed to be a drop-in replacement for Requests that is no longer under feature freeze. Supports HTTP/1.1, HTTP/2 and HTTP/3 out-of-the-box without breaking a sweat!",
      "mainEntityOfPage": {
        "@type": "WebPage",
        "@id": "https://niquests.readthedocs.io/"
      },
      "keywords": "Python Requests, Requests, Niquests, Python Niquests, http2, http3, ocsp, certificate revocation, multiplexed, Requests alternative, Python Requests Documentation, Async HTTP Python, HTTP Freethreaded",
      "articleBody": "Niquests is an elegant and simple HTTP library for Python, built for human beings. It is designed to be a drop-in replacement for Requests that is no longer under feature freeze. Supports HTTP/1.1, HTTP/2 and HTTP/3 out-of-the-box without breaking a sweat!"
    }
    </script>
{%- endblock -%}